#include <general.h>
#include "decode.h"
#include "opcodes.h"
#include "registers.h"

extern struct cpu_reg reg;
extern cpu_instruction dcpu_instr[64];
extern cpu_instruction dcpu_special[64];
extern cpu_instruction dcpu_regimm[32];
extern cpu_instruction dcpu_cop0[8];
extern cpu_instruction dcpu_cop1[32];
extern cpu_instruction dcpu_cop1_s[64];
extern cpu_instruction dcpu_cop1_d[64];
extern cpu_instruction dcpu_cop1_wl[64];

extern char* reg_names[];
extern uint32 op;

int dCPU_UNIMPLEMENTED(void){
  fprintf(stderr,"Unimplemented opcode (disasm) %d\n",opcode(op));
  return -1;
}

int dCPU_SPECIAL_UNIMPLEMENTED(void){
  fprintf(stderr,"Unimplemented special opcode (disasm) %d\n",special(op));
  return -2;
}

int dCPU_REGIMM_UNIMPLEMENTED(void){
  fprintf(stderr,"Unimplemented regimm opcode (disasm) %d\n",rt(op));
  return -3;
}

int dCPU_COP0_UNIMPLEMENTED(void){
  fprintf(stderr,"Unimplemented cop0 opcode (disasm) %d\n",special(op));
  return -4;
}

int dCPU_COP1_UNIMPLEMENTED(void){
  fprintf(stderr,"Unimplemented cop1 opcode (disasm) %d,%d\n",opcode(op)>>26,base(op),special(op));
  return -4;
}

int dCPU_LUI(void) {
 printf("lui     $%s $0x%x\n", reg_names[rt(op)], immediate(op));
}

int dCPU_ORI(void) {
 printf("ori	$%s $%s $0x%x\n", reg_names[rs(op)], reg_names[rt(op)], immediate(op));
}

int dCPU_ADDIU(void) {
 printf("addiu	$%s $%s $0x%x\n", reg_names[rt(op)], reg_names[rs(op)], immediate(op));
}

int dCPU_BEQ(void) {
 printf("beq	$%s $%s $0x%x\n", reg_names[rs(op)], reg_names[rt(op)], immediate(op));
}

int dCPU_SPECIAL(void) {
 dcpu_special[special(op)]();
}

int dCPU_SPECIAL_SLL(void) {
 printf("sll	$%s $%s $0x%x\n", reg_names[rs(op)], reg_names[rt(op)], sa(op));
}

int dCPU_SPECIAL_SLTU(void) {
 printf("sltu	$%s $%s $%s\n", reg_names[rs(op)], reg_names[rt(op)], reg_names[rd(op)]);
}

int dCPU_SPECIAL_BREAK(void) {
 printf("break	$0x%x\n", code(op));
}

int dCPU_SPECIAL_ADDU(void) {
 printf("addu    $%s $%s $%s\n", reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_JR(void) {
 printf("jr      $%s\n",reg_names[rs(op)]);
}

int dCPU_SPECIAL_MULT(void) {
 printf("mult    $%s $%s\n",reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_MFLO(void) {
 printf("mflo    $%s\n",reg_names[rd(op)]);
}

int dCPU_SPECIAL_SUB(void) {
 printf("sub     $%s $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_AND(void) {
 printf("and     $%s $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_SRA(void) {
 printf("sra     $%s $%s $0x%x\n",reg_names[rd(op)],reg_names[rt(op)],sa(op));
}

int dCPU_SPECIAL_SUBU(void) {
 printf("subu    $%s $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_OR(void) {
 printf("or	$%s $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_SRL(void) {
 printf("srl	$%s $%s $%x\n",reg_names[rd(op)],reg_names[rt(op)],sa(op));
}

int dCPU_SPECIAL_SLT(void) {
 printf("slt	$%s $%s $%s\n",reg_names[rs(op)], reg_names[rt(op)], reg_names[rd(op)]);
}

int dCPU_SPECIAL_JALR(void) {
 printf("jalr    $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)]);
}

int dCPU_SPECIAL_MFHI(void) {
 printf("mfhi	$%s\n",reg_names[rd(op)]);
}

int dCPU_SPECIAL_XOR(void) {
 printf("xor     $%s $%s $%s\n",reg_names[rd(op)],reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_DIV(void) {
 printf("div	$%s $%s\n",reg_names[rs(op)],reg_names[rt(op)]);
}

int dCPU_SPECIAL_MTLO(void) {
 printf("mtlo	$%s\n", reg_names[rs(op)]);
}

int dCPU_SPECIAL_MTHI(void) {
 printf("mthi	$%s\n", reg_names[rs(op)]);
}

int dCPU_SPECIAL_DSRLV(void) {
 printf("dsrlv	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DSLL32(void) {
 printf("dsll32	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], sa(op));
}

int dCPU_SPECIAL_DSRA32(void) {
 printf("dsra32 	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], sa(op));
}

int dCPU_SPECIAL_DDIVU(void) {
 printf("ddivu	$%s $%s\n", reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DSLLV(void) {
 printf("dsllv	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], reg_names[rs(op)]);
}

int dCPU_SPECIAL_DDIV(void) {
 printf("ddiv	$%s $%s\n", reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DMULTU(void) {
 printf("dmultu	$%s $%s\n", reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DADDU(void) {
 printf("daddu	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DSRAV(void) {
 printf("dsrav	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], reg_names[rs(op)]);
}

int dCPU_SPECIAL_SLLV(void) {
 printf("sllv	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], reg_names[rs(op)]);
}

int dCPU_SPECIAL_NOR(void) {
 printf("nor	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_DIVU(void) {
 printf("divu	$%s $%s\n", reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_SRLV(void) {
 printf("srlv	$%s $%s $%s\n", reg_names[rd(op)], reg_names[rt(op)], reg_names[rs(op)]);
}

int dCPU_SPECIAL_ADD(void) {
 printf("add	 $%s $%s $%s\n", reg_names[rd(op)], reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_SPECIAL_MULTU(void) {
 printf("multu	$%s $%s\n", reg_names[rs(op)], reg_names[rt(op)]);
}
				
int dCPU_BNE(void) {
 printf("bne     $%s $%s $0x%x\n", reg_names[rs(op)], reg_names[rt(op)],immediate(op));
}

int dCPU_SW(void) {
 printf("sw      $%s $0x%x($%s)(=0x%x)\n",reg_names[rt(op)],immediate(op),reg_names[rs(op)],(int16)immediate(op)+reg.gpr[rs(op)]);
}

int dCPU_JAL(void) {
 printf("jal     $0x%x\n",target(op)<<2);
}

int dCPU_J(void) {
 printf("j       $0x%x\n",target(op)<<2);
}

int dCPU_LW(void) {
 printf("lw	$%s $0x%x($%s)(=0x%x)\n",reg_names[rt(op)],immediate(op),reg_names[rs(op)],(int16)immediate(op)+reg.gpr[rs(op)]);
}

int dCPU_LWC1(void) {
 // eek co-processor stuff
 printf("lwc1    $%s $0x%x($%s)\n",reg_names[rt(op)],immediate(op),reg_names[rs(op)]);
}

int dCPU_COP1(void) {
  dcpu_cop1[base(op)]();
}

int dCPU_COP1_S(void) {
  dcpu_cop1_s[special(op)]();
}

int dCPU_COP1_D(void) {
  dcpu_cop1_d[special(op)]();
}

int dCPU_COP1_WL(void) {
  dcpu_cop1_wl[special(op)]();
}

int dCPU_SLTI(void) {
 printf("slti    $%s $%s $0x%x\n",reg_names[rt(op)],reg_names[rs(op)],immediate(op));
}

int dCPU_BNEL(void) {
 printf("bnel    $%s $%s $0x%x\n",reg_names[rs(op)],reg_names[rt(op)],immediate(op));
}

int dCPU_SB(void) {
 printf("sb      $%s $0x%x($%s)(=0x%x)\n",reg_names[rt(op)],immediate(op),reg_names[rs(op)],(int16)immediate(op)+reg.gpr[rs(op)]);
}

int dCPU_SLTIU(void) {
 printf("sltiu   $%s $%s $0x%x\n",reg_names[rt(op)],reg_names[rs(op)],immediate(op));
}

int dCPU_LBU(void) {
 printf("lbu     $%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}




int dCPU_REGIMM(void) {
 dcpu_regimm[rt(op)]();
}

int dCPU_REGIMM_BGEZ(void) {
 printf("bgez    $%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_REGIMM_BLTZ(void) {
 printf("bltz	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_REGIMM_BLTZL(void) {
 printf("bltzl	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_REGIMM_BGEZL(void) {
 printf("bgezl	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_REGIMM_BGEZAL(void) {
 printf("bgezal	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}			
	
int dCPU_ANDI(void) {
 printf("andi	$%s $%s $0x%x\n", reg_names[rt(op)], reg_names[rs(op)], immediate(op));
}

int dCPU_SH(void) {
 printf("sh	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op),reg_names[rs(op)]);
}

int dCPU_LHU(void) {
 printf("lhu	$0x%x($%s) $%s\n", immediate(op), reg_names[rs(op)], reg_names[rt(op)]);
}

int dCPU_XORI(void) {
 printf("xori    $%s $%s $0x%x\n",reg_names[rt(op)],reg_names[rt(op)],immediate(op));
}

int dCPU_SWL(void) {
 printf("swl	$%s %0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_BLEZ(void) {
 printf("blez    $%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_COP0(void) {
  dcpu_cop0[special(op)]();
}

int dCPU_SD(void) {
 printf("sd	$%s $0x%x($%s)(=0x%x)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)],(int16)immediate(op)+reg.gpr[rs(op)]);
}

int dCPU_LD(void) {
 printf("ld	$%s $0x%x($%s)(=0x%x)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)],(int16)immediate(op)+reg.gpr[rs(op)]);
}

int dCPU_SDC1(void) {
 printf("sdc1	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);	
}

int dCPU_ADDI(void) {
 printf("addi	$%s $%s $0x%x\n",reg_names[rt(op)],reg_names[rs(op)],immediate(op));
}

int dCPU_LDC1(void) {
 printf("ldc1	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_DADDIU(void) {
 printf("daddiu	$%s $%s $0x%x\n", reg_names[rt(op)], reg_names[rs(op)], immediate(op));
}

int dCPU_LH(void) {
 printf("lh	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}	

int dCPU_BGTZ(void) {
 printf("bgtz	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_LB(void) {
 printf("lb	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_SWC1(void) {
 printf("swc1	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_BEQL(void) {
 printf("beql	$%s $%s $0x%x\n", reg_names[rs(op)], reg_names[rt(op)],  immediate(op));
}

int dCPU_LWL(void) {
 printf("lwl	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_LWR(void) {
 printf("lwr	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_SWR(void) {
 printf("swr	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_BLEZL(void) {
 printf("blezl	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_BGTZL(void) {
 printf("bgtzl	$%s $0x%x\n",reg_names[rs(op)],immediate(op));
}

int dCPU_COP0_CACHE(void) {
 // eek more!
 printf("cop0_cache	 $%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_COP1X(void) {
 printf("unimplemented cop1x instruction\n");
}		

int dCPU_SWC2(void) {
 printf("swc2 	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}
	
int dCPU_LWC2(void) {
 printf("lwc2	$%s $0x%x($%s)\n", reg_names[rt(op)], immediate(op), reg_names[rs(op)]);
}

int dCPU_COP2(void) {
 // eek coproc stuff
 printf("cop2	$0x%x\n",target(op));
}

int dCPU_DADDI(void) {
 printf("daddi	$%s $%s $0x%x\n", reg_names[rt(op)], reg_names[rs(op)], immediate(op));
}

int dCPU_COP0_MOVE(void) {
 if(base(op))
   // MTC0
    printf("mtc0 $%s $0x%x\n",reg_names[rt(op)],rd(op)); 
 else //MFC0
   printf("mfc0 $%s $0x%x\n",reg_names[rt(op)],rd(op));
}

int dCPU_COP0_TLBWI(void) {
 printf("tlbwi\n");
}

int dCPU_COP0_TLBP(void) {
 printf("tlbp\n");
}

int dCPU_COP1_CFC1(void) {
  printf("cfc1 $%s $%s\n",reg_names[rt(op)],reg_names[rd(op)]);
}

int dCPU_COP1_CTC1(void) {
  printf("ctc1 $%s $%d\n",reg_names[rt(op)],fs(op));
}

int dCPU_COP0_ERET(void) {
  printf("eret\n");
}
