struct cpu_reg {
	long long gpr[32];  //r4300i registers
	long long gpr0[32]; //COP0 registers
	long long gpr1[32]; //COP1 registers
	unsigned int cr0[32];	     //COP0 Control registers (do they exist?)
	unsigned int cr1[32];	     //COP1 Control registers (only 0 and 31 exist)
				//But i'd rather trade memory in for speed than
				//vice versa
	long long pc;		//Program counter
	long long HI;		//High 64 bits of multiply/divide result
	long long LO;		//Low 64 bits of multiply/divide result
	char LLbit;		//Load/Link bit, as a char :O)
	int64 CPUdelayPC;
	uint32 CPUdelay;
};

#define MAXTLB 32

struct tlb_struct
{
    uint32  valid;
    uint32  EntryHi;
    uint32  EntryLo1;
    uint32  EntryLo0;
    uint64  PageMask;
    uint32  LoCompare;
    uint32  MyHiMask;
};

extern struct tlb_struct        TLB[MAXTLB];

#define COP0_STATUS	reg.gpr0[12]
#define COP0_CAUSE	reg.gpr0[13]
#define COP0_EPC	reg.gpr0[14]
#define COP0_ERROR_EPC	reg.gpr0[30]

// COP0_STATUS bitmasks etc
#define COP0_STATUS_IE		( COP0_STATUS & 1 )
#define COP0_STATUS_EXL		( COP0_STATUS & 2 )
#define COP0_STATUS_ERL		( COP0_STATUS  & 4 )
#define COP0_STATUS_KSU		(( COP0_STATUS & 24)>>3)
#define COP0_STATUS_UX		( COP0_STATUS & 32 )
#define COP0_STATUS_SX		( COP0_STATUS & 64 )
#define COP0_STATUS_KX		( COP0_STATUS & 128 )
#define COP0_STATUS_IMF		(( COP0_STATUS & 65280UL ) >> 8) // add subfileds if needed
#define COP0_STATUS_DS		(( COP0_STATUS & 33488896UL ) >> 16 ) // ditto
#define COP0_STATUS_RE		( COP0_STATUS & 33554432UL )
#define COP0_STATUS_FR		( COP0_STATUS & 67108864UL )
#define COP0_STATUS_RP		( COP0_STATUS & 134217728UL )
#define COP0_STATUS_CU		(( COP0_STATUS & 134217728UL )>> 28 )

// bitmasks useful for setting stuff
#define COP0_STATUS_IE_BM	( 1 )
#define COP0_STATUS_EXL_BM	( 2 )
#define COP0_STATUS_ERL_BM	( 4 )
#define COP0_STATUS_KSU_BM	( 24 )
#define COP0_STATUS_UX_BM	( 32 )
#define COP0_STATUS_SX_BM	( 64 )
#define COP0_STATUS_KX_BM	( 128UL )
#define COP0_STATUS_IMF_BM	( 65280UL )
#define COP0_STATUS_DS_BM	( 33488896UL )
#define COP0_STATUS_RE_BM	( 33554432UL )
#define COP0_STATUS_FR_BM	( 67108864UL )
#define COP0_STATUS_RP_BM	( 134217728UL )
#define COP0_STATUS_CU_BM	( 134217728UL  )

#define COP0_CAUSE_BD_BM	( 0x80000000 )
#define COP0_CAUSE_EXCCODE_BM	( 0x7c )
